@@ -1,7 +1,13 @@
 """pysolarmanv5.py"""
+import queue
 import struct
 import socket
 import logging
+import select
+from threading import Thread, Event
+from multiprocessing import Queue
+from typing import Any
+from random import randrange
 
 from umodbus.client.serial import rtu
 
@@ -51,6 +57,13 @@ class PySolarmanV5:
     :type socket: Socket, optional
     :raises NoSocketAvailableError: If no network socket is available
 
+    .. versionadded:: v2.5.0
+
+    :param auto_reconnect: Activates the auto-reconnect functionality.
+        PySolarmanV5 will try to keep the connection open. The default is False.
+        Not compatible with custom sockets.
+    :type auto_reconnect: Boolean, optional
+
     .. deprecated:: v2.4.0
 
     :param verbose: Enable verbose logging, defaults to False. Use **logger**
@@ -83,22 +96,29 @@ class PySolarmanV5:
         self.verbose = kwargs.get("verbose", False)
         self.socket_timeout = kwargs.get("socket_timeout", 60)
         self.v5_error_correction = kwargs.get("error_correction", False)
+        self.sequence_number = None
 
         if self.verbose:
             self.log.setLevel("DEBUG")
 
         self._v5_frame_def()
 
-        self.sock = kwargs.get("socket", self._create_socket())
-        if self.sock is None:
-            raise NoSocketAvailableError("No socket available")
+        self.sock: socket.socket = None  # noqa
+        self._poll: select.poll = None  # noqa
+        self._sock_fd: int = None  # noqa
+        self._auto_reconnect = False
+        self._data_queue: Queue = None  # noqa
+        self._data_wanted: Event = None  # noqa
+        self._reader_exit: Event = None  # noqa
+        self._reader_thr: Thread = None  # noqa
+        self._socket_setup(kwargs.get("socket"), kwargs.get("auto_reconnect", False))
 
     def _v5_frame_def(self):
         """Define and construct V5 request frame structure."""
         self.v5_start = bytes.fromhex("A5")
         self.v5_length = bytes.fromhex("0000")  # placeholder value
         self.v5_controlcode = struct.pack("<H", 0x4510)
-        self.v5_serial = bytes.fromhex("0000")
+        self.v5_serial = bytes.fromhex("0000")  # placeholder value
         self.v5_loggerserial = struct.pack("<I", self.serial)
         self.v5_frametype = bytes.fromhex("02")
         self.v5_sensortype = bytes.fromhex("0000")
@@ -121,7 +141,22 @@ class PySolarmanV5:
         checksum = 0
         for i in range(1, len(frame) - 2, 1):
             checksum += frame[i] & 0xFF
-        return int((checksum & 0xFF))
+        return int(checksum & 0xFF)
+
+    def _get_next_sequence_number(self):
+        """Get the next sequence number for use in outgoing packets
+
+        If ``sequence_number`` is None, generate a random int as initial value.
+
+        :return: Sequence number
+        :rtype: int
+
+        """
+        if self.sequence_number is None:
+            self.sequence_number = randrange(0x01, 0xFF)
+        else:
+            self.sequence_number = (self.sequence_number + 1) & 0xFF
+        return self.sequence_number
 
     def _v5_frame_encoder(self, modbus_frame):
         """Take a modbus RTU frame and encode it in a V5 data logging stick frame
@@ -134,6 +169,7 @@ class PySolarmanV5:
         """
 
         self.v5_length = struct.pack("<H", 15 + len(modbus_frame))
+        self.v5_serial = struct.pack("<H", self._get_next_sequence_number())
 
         v5_header = bytearray(
             self.v5_start
@@ -173,11 +209,12 @@ class PySolarmanV5:
 
         1) V5 start and end are correct (``0xA5`` and ``0x15`` respectively)
         2) V5 checksum is correct
-        3) V5 data logger serial number is correct (in most (all?) instances the
+        3) V5 outgoing sequence number has been echoed back to us (byte 5)
+        4) V5 data logger serial number is correct (in most (all?) instances the
            reply is correct, but request can obviously be incorrect)
-        4) V5 control code is correct (``0x1510``)
-        5) v5_frametype contains the correct value (``0x02`` in byte 11)
-        6) Modbus RTU frame length is at least 5 bytes (vast majority of RTU
+        5) V5 control code is correct (``0x1510``)
+        6) v5_frametype contains the correct value (``0x02`` in byte 11)
+        7) Modbus RTU frame length is at least 5 bytes (vast majority of RTU
            frames will be >=6 bytes, but valid 5 byte error/exception RTU frames
            are possible)
 
@@ -204,6 +241,8 @@ class PySolarmanV5:
             raise V5FrameError("V5 frame contains invalid start or end values")
         if v5_frame[frame_len - 2] != self._calculate_v5_frame_checksum(v5_frame):
             raise V5FrameError("V5 frame contains invalid V5 checksum")
+        if v5_frame[5] != self.sequence_number:
+            raise V5FrameError("V5 frame contains invalid sequence number")
         if v5_frame[7:11] != self.v5_loggerserial:
             raise V5FrameError("V5 frame contains incorrect data logger serial number")
         if v5_frame[3:5] != struct.pack("<H", 0x1510):
@@ -228,13 +267,94 @@ class PySolarmanV5:
 
         """
         self.log.debug("SENT: " + data_logging_stick_frame.hex(" "))
-
+        if not self._reader_thr.is_alive():
+            raise NoSocketAvailableError("Connection already closed.")
         self.sock.sendall(data_logging_stick_frame)
-        v5_response = self.sock.recv(1024)
+        self._data_wanted.set()
+        try:
+            v5_response = self._data_queue.get(timeout=self.socket_timeout)
+            if v5_response == b"":
+                raise NoSocketAvailableError("Connection closed on read")
+            self._data_wanted.clear()
+        except TimeoutError:
+            raise
 
         self.log.debug("RECD: " + v5_response.hex(" "))
         return v5_response
 
+    def _data_receiver(self):
+        self._poll.register(self.sock.fileno(), select.POLLIN)
+        while True:
+            events = self._poll.poll(500)
+            if self._reader_exit.is_set():
+                return
+            for event in events:
+                # We are registered only for inbound data on a single socket,
+                # so there is no need to check the (fileno, mask) tuples
+                data = self.sock.recv(1024)
+                if data == b"":
+                    self.log.debug(f"[POLL] Socket closed. Reader thread exiting.")
+                    if self._data_wanted.is_set():
+                        try:
+                            self._data_queue.put_nowait(data)
+                        except queue.Full:
+                            pass
+                    self._reconnect()
+                    return
+                elif data.startswith(b"\xa5\x01\x00\x10G"):
+                    # Frame with control code 0x4710 - Counter frame
+                    self.log.debug(f'[{self.serial}] COUNTER: {data.hex(" ")}')
+                    continue
+                if self._data_wanted.is_set():
+                    self._data_queue.put(data, timeout=self.socket_timeout)
+                else:
+                    self.log.debug("[POLL-DISCARDED] RECD: " + data.hex(" "))
+
+    def _reconnect(self):
+        """
+        Reconnect to the data logger if needed
+        """
+        if self._reader_thr.is_alive():
+            try:
+                self.sock.send(b"")
+                self.sock.close()
+            except OSError:
+                pass
+            self._reader_exit.set()
+        if self._auto_reconnect:
+            self.log.debug(
+                f"Auto-Reconnect enabled. Trying to establish a new connection"
+            )
+            self._poll.unregister(self._sock_fd)
+            self.sock = self._create_socket()
+            if self.sock:
+                self._sock_fd = self.sock.fileno()
+                self._reader_exit.clear()
+                self._reader_thr = Thread(target=self._data_receiver, daemon=True)
+                self._reader_thr.start()
+                self.log.debug(f"Auto-Reconnect successful.")
+            else:
+                self.log.debug(f"No socket available! Reconnect failed.")
+        else:
+            self.log.debug("Auto-Reconnect inactive.")
+
+    def disconnect(self) -> None:
+        """
+        Disconnect the socket and set a signal for the reader thread to exit
+
+        :return: None
+
+        """
+        self._data_wanted.clear()
+        try:
+            self.sock.send(b"")
+            self.sock.close()
+        except OSError:
+            pass
+        self._reader_exit.set()
+        self._reader_thr.join(0.5)
+        self._poll.unregister(self._sock_fd)
+
     def _send_receive_modbus_frame(self, mb_request_frame):
         """Encodes mb_frame, sends/receives v5_frame, decodes response
 
@@ -272,6 +392,22 @@ class PySolarmanV5:
             return None
         return sock
 
+    def _socket_setup(self, sock: Any, auto_reconnect: bool):
+        """Socket setup method"""
+        if isinstance(sock, socket.socket) or sock is None:
+            self.sock = sock if sock else self._create_socket()
+            if self.sock is None:
+                raise NoSocketAvailableError("No socket available")
+            self._poll = select.poll()
+            self._sock_fd = self.sock.fileno()
+            self._auto_reconnect = False if sock else auto_reconnect
+            self._data_queue = Queue(maxsize=1)
+            self._data_wanted = Event()
+            self._reader_exit = Event()
+            self._reader_thr = Thread(target=self._data_receiver, daemon=True)
+            self._reader_thr.start()
+            self.log.debug(f"Socket setup completed... {self.sock}")
+
     @staticmethod
     def twos_complement(val, num_bits):
         """Calculate 2s Complement
